#!/usr/bin/perl -w
###############################################################################
## Project:	Compiler Construction Course
## Descr:	Virtual Assembly Machine:
##              An Interpreter of a Simple Processor's Instruction Set
## Author:	Dr. Jrgen Vollmer, <Juergen.Vollmer@informatik-vollmer.de>
## $Id: vam.pl,v 1.8 2005/07/12 12:06:43 vollmer Exp $
###############################################################################

=pod

=head1 NAME

vam.pl - Virtual Assembler Machine: an interpreter for a simple processor's instruction set

=head1 SYNOPSIS

vam.pl [--help] [--manual | -M] [--verbose ] [--Version]
       [--input infile] [--output outfile]
       [--registers count] [--memory size]
       [--dump | -D]
       [--statistics]
       [srcfile...]

=head1 OPTIONS

=over

=item B<--input> I<infile>

Read the input of the interpreted program, from I<infile>.

Default: F<stdin>.

=item B<--output> I<outfile>

Write the output of the interpreted program, to I<outfile>.

Default: F<stdout>.

=item B<--memory> I<size>

Size of the avaliable memory.
I<size> is a positive number followed by B<k> (kilo) or
B<M> (mega), which specifies the size in kilo or mega memory cells
respectively.

Default: 32M

=item B<--registers> I<count>

Number of the available registers: B<R0>, B<R1>, ... B<R>I<count-1>.

Default: 32

=item B<--help>

Show this help and terminate.

=item B<--manual> B<-M>

Show the entire manual and terminate.

=item B<--verbose>

Verbose.

=item B<--Version>

Show program version.

=item B<--dump>

=item B<-D>

After parsing the source file, dump the instruction list.

=item B<--statistics>

Emit some statistical information about the program.

=back

Option names may abbreviated as long as the are unique.
Instead of --XX -XX may be used.

=head1 DESCRIPTION

B<vam.pl> is an interpreter of a simple processor's (designed by me and called
B<VAM>, Virtual Assembler Processor) instruction set. It has the "usual"
machine instructions available on "real" processors.

B<vam.pl> may be used e.g. in a compiler construction course to have a "play
ground" when generating machine code.

B<vam.pl> read the instructions to be interpreted either from F<stdin> or
the file(s) I<srcfile> given on the command line.
The I<BIOS> (Basic Input / Output System) of the processor allows
to read and write a file during runtime (see B<--input> and B<--output>).

=head1 FORMAT OF THE SOURCE FILE

The source file I<srcfile> contains the assembler instructions to be
interpreted.

A line is either empty, contains a comment or I<one> instruction or I<one>
(string or data) declaration.

The # character starts a comment up to the end of the line.

The instruction and registers names are case insensitive.
Labels and string literals are case sensitive.

=cut

use warnings;
use strict;
use English;
use locale;
use Getopt::Long;
use Pod::Usage;
use File::Basename qw(dirname basename);

our $CMD            = basename ($0);
our ($VERSION)	    = '1.0';
our ($VERSION_DATE) = 'Date: 2004/01/16 13:23:28 $' =~ '(\d*/\d*/\d*)';

our ($option_verbose,  $option_version,
     $option_output,   $option_input,
     $option_dump,     $option_statistics,
     $max_registers,   $max_memory
     );

###############################################################################
##  the main data structures
###############################################################################

our @R;  # array of Registers
         # R0 = instruction counter
         # R1 = stack pointer
         # R2 = free_memory-pointer

our @I;  # Array of of hash of instructions to be interpreted
         # $I[<address>] = {file:    string  // the name of the source file
         #                  line:    int     // line number in the source
         #                  label:   string  // the label of an instruction
         #                                   // or "" if no label is given
	 #                  address: int     // the instruction's address
         #                  text:    string  // instruction text
         # $M<address> referes to the instruction represented by $I<address>.

our @M;  # the Memory holding values and instructions.
         # $M[<address>] =
         #   ordinary values:
         #          = <integer value>
         #     or   = <real value>
         # If a memory cell contains an instruction, $M[address] stores
         # an array
         #          = [<instruction>, <address>, <argument1>, <argument2>, ...]
         # where <instruction> is a reference to a routine "vam_<instr>"
         # which implements the VAM instruction <instr>.
         # The interprete() - routine
         #  - fetches the instruction referenced by the instruction
         #    counter (R0)
         #  - calls the interpreation routine:
         #    vam_<instr> ([<instruction>, <address>, <argument1>,
         #                 <argument2>, ...])
         #  - that routine update the instruction counter.
         # The interpretation stops if
         #  - M[R0] does not specify an instruction, or
         #  - R0 < 1, or
         #  - R0 > $last_instr, which is an error.
         # The "end" instructions sets R0 to 0, which indicates
         # termination.

our %A;  # Hash of instruction arguments:
         # Checking the arguments of instruction is done in an generic way,
         # see &get_args():
         # $A{<instr>} = "..."
	 # where ... is a sequence of the following characters:
         #  R : register
         #  L : instruction label
         #  S : string label
         #  D : data label
         #  I : integer constant
         #  F : float constant
         # e.g.:
         # -  $A{add_i} = "RRR"
         #    which means: the add_i instruction takes three register
         #                 arguments.
         # -  $A{cload_i} = "RI"
         #    which means: the cload_i instruction takes as first argument
         #    a register and as second one an integer literal

our $last_instr;
         # address of the last instruction;
         # Memory[0] is undefined
         # the first valid instruction has address 1

our $malloc_start;
         # address first fee memory

our %S;  # the "String" pool
         # $s{<label>} = "..."
         # Strings are not stored in the @M array.

our %L;  # The mapping instruction Label --> address
         # $L{<label>} = <address>

our %D;  # The mapping data Label --> address
         # $D{<label>} = <address>

###############################################################################
## some helpers
###############################################################################

sub error
# emit a general error message and die
{
    die "$CMD: ", @_, "\n";
}

###############################################################################

sub syntax
# emit a syntax error message and die
# Arguments:  1     : $instr (a hash value of @I)
#             2,... : message
{
    my $instr = shift;
    die $instr->{file} . " line " . $instr->{line} .": ", @_, "\n";
}

sub syntax2
# Arguments:  1     : $file
#             2     : $line
#             2,... : $message
{
    my $file = shift;
    my $line = shift;
    die "$file line $line: ", @_, "\n";
}

#############################################################################

sub cm($)
# check memory access, if the memory is undefined, store a 0
# Arguments: 1: the address
# returns:   the address
{
    my $adr = shift;
    (0 < $adr && $adr < $max_memory) ||
	syntax ($I[$R[0]], "Illegal memory address: $adr");
    $M[$adr] = 0 if (!defined $M[$adr]);
    return $adr;
}

#############################################################################

sub dump_memory
# dump content of memory and string pool to stdout
{
    printf "Instructions\n";
    printf "[file line nr] address label : instruction\n";
    printf "-----------------------------+------------\n";
    for my $i (@I) {
	printf "[%s %4d] %4d %-10s: %s\n",
	$i->{file}, $i->{line}, $i->{address}, $i->{label}, $i->{text};
    }

    printf "\nData pool\n";
    printf "%-10s : %8s : %6s\n", "label", "address", "size";
    printf "%s-:-%s-:-%s\n", "-"x10,  "-"x8, "-"x6;
    for my $label (sort {$a cmp $b} keys %D) {
	printf "%-10s : %8d : %6d\n", $label,
	        $D{$label}{address},
	        $D{$label}{size}, ;
    }

    printf "\nString pool\n";
    printf "%-10s: %s\n", "label", "text";
    printf "%10s+-------------\n", "-"x10;
    for my $label (sort {$a cmp $b} keys %S) {
	printf "%-10s: %s\n", $label, $S{$label};
    }
}

#############################################################################

=pod

=head1 ARCHITECTURE OF THE PROCESSOR VAM

The VAM processor has a usual I<von-Neumann> architecture with registers,
memory, an ALU (arithmetic logical unit) etc.

The processor knows two data types: (signed) integers and floating point numbers.
Memory addresses are positive integer values.

The number of registers may be specified by the B<--registers> command line
option and are denoted by B<R0>, B<R1>, ... A register may hold I<one> value of
a given data type at a time.

The memory is partitioned into I<memory cells>, which may hold I<one> value of
a given data type at a time.

=head2 Special Registers

The registers B<R0>, B<R1> und B<R2> are reserved and should not be used
for other purposes (even it is legal to modify them).
All other registers may be used without any restrictions.

=over

=item B<R0>

The register B<R0> is used as I<instruction counter> and set to 1 before
the program starts.

With the exception of the control flow and non-executable statements, after
performing the operation specified by an instruction, the instruction counter
B<R0> will be incremented by 1.

=item B<R1>

The register B<R1> is used as stack pointer pointing to the top of the stack
maintained by some instructions (see below). At program start B<R1> is initialized to
the value of I<high-memory>.

=item B<R2>

The register B<R2> holds the pointer to the first unused memory cell.  The the
free memory may be used manage dynamically allocated memory similar to the
I<sbrk>(2) call of the C-library.

=back

=head2 Layout of the Memory


    Address        | Description
   ----------------+-------------------------------------------------
       0           | Undefined address
   ----------------+-------------------------------------------------
       1           | The instruction to be executed at program start
      ...          | More instructions.
     <code-end>    | The textual last instruction
   ----------------+-------------------------------------------------
    <data-start>   | Global data
     ...           | ........
    <data-end>     | ........
   ----------------+-------------------------------------------------
    <malloc-start> | Start of free memory.
     ...           | ....
    <high-memory>-1| Last memory cell, initial top of stack
   ----------------+-------------------------------------------------

The free memory "grows" towards larger address, while the stack grows towards
small addresses.

=head2 VAM Runtime Errors

The VAM emits under the following conditions an error message and terminates:

=over

=item

* VAM does not emit an error if the two areas overlap.

=item

* VAM emits an error if the memory at address 0 or an address >= I<high-memory>
is accessed.

=item

* VAM emits an error, if a memory cell not containing an instruction should be
read as instruction and executed.

=item

* Division by 0 causes an error.

=back

=cut

###############################################################################
## The workers:
## For each instruction there is procedure which interprets it.
## The routine is called by the "interprete" routine.
## The arguments are:
## $_[ ] :
##    0    the reference to the routine
##    1    the reference to the instruction's hash
##    2    the first operand of the instruction, optional
##    3    the second    -"-                   , optional
##    4    the third     -"-                   , optional
## Each routine updates the instruction counter.
###############################################################################

=pod

=head1 MACHINE INSTRUCTIONS

In the sequel the instruction are declared, which the processor is able to
process.
In the description below I<r>, I<r1>, I<r2>, I<r3> specify any processor
register B<R0>, B<R1>, ..., B<R>I<max_registers-1>, even the reserved ones.
If for an instruction several registers may be specified, the same register may
be used several times.

The notion B<M[>I<address>B<]> specifies the access of the memory cell
at the given I<address>.

=head2 Labels

There are three kinds of named entities, named by so called I<label>:

=over

=item 1. Data declaration:

I<label> B<:> B<DATA> <count>

=item 2. String declaration

I<label> B<:> B<">character stringB<">

=item 3. Instruction

I<label> B<:> I<one-instruction>

=back

A I<label> consists of a letter followed by letters, digits or an underscore
character. The labels of each entity class (data, string, instruction)
reside in a separate namespace and must be unique in that namespace.

The string and data declarations are non-executable instructions, therefore the
instruction counter is not modified. They may occur everywhere in the source
file.

Memory cells holding global data must be reserved and a label must be given be
given.  This declaration reserves I<count> memory cells. The address of the
first memory cell may be accessed using the name I<label>.

A string declaration declares a name for a character string used in the
B<write_s> instruction.  The two characters "\n" represent a newline. The two
characters "\t" represent a tabulator.

A control flow statement may use an instruction I<label> to specify the next
statement to be executed.


Some instructions have a suffix in their name.  The suffix B<_f> indicates
floating point and B<_i> integer operation. The suffix B<_c> indicates that the
instruction takes a literal integer argument, B<_r> a register is
used and B<_l> a label.

A prefix B<i> indicates some kind of indirection.

=head2 Control Flow Instructions

=over

=cut

###############################################################################

=pod

=item B<nop>

No Operation, do nothing and continue execution with the next instruction.
If a label declaration is not followed by an instruction,
a B<nop> is assumed.

=cut

$A{"nop"} = "";
sub vam_nop {$R[0]++;}

###############################################################################

=pod

=item B<end>

Stop the processor and terminate the interpretation.

=cut

$A{end} = "";
sub vam_end {$R[0] = 0;}

###############################################################################

=pod

=item B<goto> I<label>

=item B<igoto> I<r>

Continue execution with the instruction named by I<label>, or at the address
stored in register I<r>.

=cut

$A{goto}  = "L"; sub vam_goto  {$R[0] = $_[2];}
$A{igoto} = "R"; sub vam_igoto {$R[0] = $R[$_[2]];}

###############################################################################

=pod

=item B<call>  I<r1> B<,> I<label>

=item B<icall> I<r1> B<,> I<r2>

Continue execution with the instruction named by I<label> (or the address
stored in register I<r2>) and push the address of the statement following the
B<call> onto the stack pointed to by I<r2>.

I<r1> = I<r1> - 1; Memory[I<r1>] = B<R0> + 1; B<R0> = I<label>

I<r1> = I<r1> - 1; Memory[I<r1>] = B<R0> + 1; B<R0> = I<r2>

=cut

$A{call}   = "RL";
sub vam_call  {$M[cm(--$R[$_[2]])] = $R[0] + 1; $R[0] = $_[3];}

$A{icall}  = "RR";
sub vam_icall {$M[cm(--$R[$_[2]])] = $R[0] + 1; $R[0] = $R[$_[3]];}

###############################################################################

=pod

=item B<return> I<r>

Get address from the top of the stack I<r>, pop it and continue execution
at the instruction specified by that address.

B<R0> = Memory[I<r>]; I<r> = I<r> + 1;

=cut

$A{return}  = "R";
sub vam_return {$R[0] = $M[cm($R[$_[2]])]; $R[$_[2]]++;}

###############################################################################

=pod

=back

=head2 Conditional Control Flow Instructions

If register I<r> holds the shown condition, then continue execution with the
instruction specified by I<label>, else continue with the textual next
instruction.

The registers should contain an integer value.

See also B<cmp_i> and B<cmp_f>.

=over

=item B<iflt>   I<r> B<,> I<label>

less then:     I<r> <  0

=item B<ifle>   I<r> B<,> I<label>

less equal:    I<r> <= 0

=item B<ifeq>   I<r> B<,> I<label>

equal:         I<r> =  0

=item B<ifne>   I<r> B<,> I<label>

not equal:     I<r> != 0

=item B<ifgt>   I<r> B<,> I<label>

greater then:  I<r> >  0

=item B<ifge>   I<r> B<,> I<label>

greater equal: I<r> >= 0

=item B<iftrue> I<r> B<,> I<label>

true-value:    I<r> != 0

=item B<iftrue> I<r> B<,> I<label>

false-value:   I<r> == 0

=back

=cut

$A{iflt}   = $A{ifle}    =
$A{ifeq}   = $A{ifne}    =
$A{ifgt}   = $A{ifge}    =
$A{iftrue} = $A{iffalse} = "RL";
sub vam_iflt    {$R[0] = ($R[$_[2]] <  0) ? $_[3] : $R[0]+1;}
sub vam_ifle    {$R[0] = ($R[$_[2]] <= 0) ? $_[3] : $R[0]+1;}
sub vam_ifeq    {$R[0] = ($R[$_[2]] == 0) ? $_[3] : $R[0]+1;}
sub vam_ifne    {$R[0] = ($R[$_[2]] != 0) ? $_[3] : $R[0]+1;}
sub vam_ifgt    {$R[0] = ($R[$_[2]] >  0) ? $_[3] : $R[0]+1;}
sub vam_ifge    {$R[0] = ($R[$_[2]] >= 0) ? $_[3] : $R[0]+1;}
sub vam_iftrue  {$R[0] = ($R[$_[2]] != 0) ? $_[3] : $R[0]+1;}
sub vam_iffalse {$R[0] = ($R[$_[2]] == 0) ? $_[3] : $R[0]+1;}

###############################################################################
###############################################################################

=pod

=head2 Arithmetic instructions

Except when noted, all registers used by an instruction must hold values of
the same datatype, which must match the type of the instruction.
Otherwise the result may be undefined.

The processing of an instructions read the source registers performs the
operation and store the result in the destination register.

There is no test over underflow or overflow of the result. A division by 0
causes an error.

=over

=cut


###############################################################################

=pod

=item B<add_i>  I<r1> B<,> I<r2> B<,> I<r3>

=item B<add_f>  I<r1> B<,> I<r2> B<,> I<r3>

Addition: I<r1> =  I<r2> +  I<r3>

=cut

$A{add_i}  = $A{add_f} = "RRR";
sub vam_add_i  {$R[$_[2]] = $R[$_[3]] + $R[$_[4]]; $R[0]++;}
sub vam_add_f  {$R[$_[2]] = $R[$_[3]] + $R[$_[4]]; $R[0]++;}

###############################################################################

=pod

=item B<sub_i> I<r1> B<,> I<r2> B<,> I<r3>

=item B<sub_f> I<r1> B<,> I<r2> B<,> I<r3>

Subtraction: I<r1> =  I<r2> - I<r3>

=cut

$A{sub_i} = $A{sub_f} = "RRR";
sub vam_sub_i {$R[$_[2]] = $R[$_[3]] - $R[$_[4]]; $R[0]++;}
sub vam_sub_f {$R[$_[2]] = $R[$_[3]] - $R[$_[4]]; $R[0]++;}

###############################################################################

=pod

=item B<mult_i> I<r1> B<,> I<r2> B<,> I<r3>

=item B<mult_f> I<r1> B<,> I<r2> B<,> I<r3>

Multiplication: I<r1> =  I<r2> * I<r3>

=cut

$A{mult_i} = $A{mult_f} = "RRR";
sub vam_mult_i {$R[$_[2]] = $R[$_[3]] * $R[$_[4]]; $R[0]++;}
sub vam_mult_f {$R[$_[2]] = $R[$_[3]] * $R[$_[4]]; $R[0]++;}

###############################################################################

=pod

=item B<div_i> I<r1> B<,> I<r2> B<,> I<r3>

=item B<div_f> I<r1> B<,> I<r2> B<,> I<r3>

Division: I<r1> =  I<r2> / I<r3>

If I<r3> == 0 an error message is emitted and the program is terminated.

=cut

$A{div_i} = $A{div_f} = "RRR";

sub vam_div_i
{
    if ($R [$_[4]] == 0) {
	syntax ($_[1], "Division by 0");
    } else {
	$R[$_[2]] = $R[$_[3]] / $R[$_[4]];
    }
    $R[0]++;
}

sub vam_div_f
{
    if ($R [$_[4]] == 0) {
	syntax ($_[1], "Division by 0");
    } else {
	$R[$_[2]] = $R[$_[3]] / $R[$_[4]];
    }
    $R[0]++;
}

###############################################################################

=pod

=item B<mod_i> I<r1> B<,> I<r2> B<,> I<r3>

Modulo: I<r1> =  I<r2> % I<r3>

If I<r3> == 0 an error message is emitted and the program is terminated.

Note: there is no floating point mod operation!

=cut

$A{mod_i} = "RRR";

sub vam_mod_i
{
    if ($R [$_[4]] == 0) {
	syntax ($_[1], "Division by 0");
    } else {
	$R[$_[2]] = $R[$_[3]] % $R[$_[4]];
    }
    $R[0]++;
}

###############################################################################

=pod

=item B<cmp_i> I<r1> B<,> I<r2> B<,> I<r3>

=item B<cmp_f> I<r1> B<,> I<r2> B<,> I<r3>

comparison: I<r1> =  I<r2> compare I<r3>

I<r1> holds an integer -1, 0, or 1 according to:

=over

=item S<-1> if I<r2> E<lt>  I<r3>

=item S< 0> if I<r2> == I<r3>

=item S< 1> if I<r2> E<gt>  I<r3>

=back

=cut

$A{cmp_i} = $A{sub_f} = "RRR";
sub vam_cmp_i {$R[$_[2]] = $R[$_[3]] <=> $R[$_[4]]; $R[0]++;}
sub vam_cmp_f {$R[$_[2]] = $R[$_[3]] <=> $R[$_[4]]; $R[0]++;}

###############################################################################

=pod

=item B<add_c>  I<r1> B<,> I<r2> B<,> I<value>

=item B<sub_c>  I<r1> B<,> I<r2> B<,> I<value>

=item B<mult_c> I<r1> B<,> I<r2> B<,> I<value>

=item B<div_c>  I<r1> B<,> I<r2> B<,> I<value>

=item B<mod_c>  I<r1> B<,> I<r2> B<,> I<value>

=item B<cmp_c>  I<r1> B<,> I<r2> B<,> I<value>

Integer arithmetic / comparison with an (signed) integer constant:

I<r1> =  I<r2> <operand>  I<value>

=cut

$A{add_c} = $A{sub_c} = $A{mult_c} = $A{div_c} = $A{mod_c} = $A{cmp_c} = "RRI";
sub vam_add_c  {$R[$_[2]] = $R[$_[3]] + $_[4];     $R[0]++;}
sub vam_sub_c  {$R[$_[2]] = $R[$_[3]] - $_[4];     $R[0]++;}
sub vam_mult_c {$R[$_[2]] = $R[$_[3]] * $_[4];     $R[0]++;}
sub vam_div_c
{
    if ($_[4] == 0) {
	syntax ($_[1], "Division by 0");
    } else {
	$R[$_[2]] = $R[$_[3]] / $_[4];
    }
    $R[0]++;
}
sub vam_mod_c
{
    if ($_[4] == 0) {
	syntax ($_[1], "Division by 0");
    } else {
	$R[$_[2]] = $R[$_[3]] % $_[4];
    }
    $R[0]++;
}
sub vam_cmp_c {$R[$_[2]] = $R[$_[3]] <=> $_[4]; $R[0]++;}

###############################################################################

=pod

=item B<f2i> I<r1> B<,> I<r2>

Type cast: I<r1> = (int) I<r2>  (in C speak)

=cut

$A{f2i} = "RR";
sub vam_f2i {$R[$_[2]] = int ($R[$_[3]]); $R[0]++;}

###############################################################################

=pod

=item B<i2f> I<r1> B<,> I<r2>

Type cast: I<r1> = (float) I<r2>  (in C speak)

=cut

$A{i2f} = "RR";
sub vam_i2f {$R[$_[2]] = $R[$_[3]] * 1.0; $R[0]++;}

###############################################################################

=pod

=item B<lshift> I<r1> B<,> I<r2> B<,> I<r3>

=item B<rshift> I<r1> B<,> I<r2> B<,> I<r3>

Left shift: I<r1> = I<r2> << I<r3>  (in C speak)

Right shift: I<r1> = I<r2> >> I<r3>  (in C speak)

All registers must hold integer values.

=cut

$A{lshift} = $A{rshift} = "RRR";
sub vam_lshift {$R[$_[2]] = $R[$_[3]] << $R[$_[4]]; $R[0]++;}
sub vam_rshift {$R[$_[2]] = $R[$_[3]] >> $R[$_[4]]; $R[0]++;}

###############################################################################

=pod

=back

=cut

###############################################################################
###############################################################################

=pod

=head2 Transport Instructions

Transport instructions move values from one place to another.

If a register specifies an address, it must hold an positive integer value.
Everything else is undefined.

=over

=cut

###############################################################################

=pod

=item B<copy> I<r1> B<,> I<r2>

Copies the content of registers: I<r1> = I<r2>

=cut

$A{copy} = "RR";
sub vam_copy {$R[$_[2]] = $R[$_[3]]; $R[0]++;}

###############################################################################

=pod

=item B<cload_i> I<r> B<,> I<value>

=item B<cload_f> I<r> B<,> I<value>

Loads the (signed) integer or floating point I<value> into the register I<r>.

=cut

$A{cload_i} = "RI";
sub vam_cload_i  {$R[$_[2]] = $_[3]; $R[0]++;}

$A{cload_f} = "RF";
sub vam_cload_f  {$R[$_[2]] = $_[3]; $R[0]++;}

###############################################################################

=pod

=item B<load> I<r1> B<,> I<r2>

=item B<load_l> I<r1> B<,> I<label>

=item B<load_c> I<r1> B<,> I<r2> B<,> I<value>

Load the content a memory cell to a register.
I<label> is a data label and I<value> is a signed integer.

I<r1> = Memory[I<r2>]

I<r1> = Memory[I<label>]

I<r1> = Memory[I<r2> + I<value>]

=cut

$A{load}   = "RR";
sub vam_load   {$R[$_[2]] = $M[cm($R[$_[3]])];        $R[0]++;}

$A{load_l} = "RD";
sub vam_load_l {$R[$_[2]] = $M[cm($_[3])];            $R[0]++;}

$A{load_c} = "RRI";
sub vam_load_c {$R[$_[2]] = $M[cm($R[$_[3]]+ $_[4])]; $R[0]++;}

###############################################################################

=pod

=item B<iload> I<r1> B<,> I<r2> B<,> I<r3>

Indirect load with register offset:
I<r1> = Memory[Memory[I<r2>] + I<r3>]

=cut

$A{iload} = "RRR";
sub vam_iload {$R[$_[2]] = $M[cm($M[cm($R[$_[3]])] + $R[$_[4]])]; $R[0]++;}

###############################################################################

=pod

=item B<iload_c> I<r1> B<,> I<r2> B<,> I<value>

Indirect load with constant offset:
I<r1> = Memory[Memory[I<r2>] + I<value>]

I<value> is an (signed) integer number.

=cut

$A{iload_c} = "RRI";
sub vam_iload_c {$R[$_[2]] = $M[cm($M[cm($R[$_[3]])] + $_[4])]; $R[0]++;}

###############################################################################

=pod

=item B<store> I<r1> B<,> I<r2>

=item B<store_l> I<label> B<,> I<r2>

=item B<store_c> I<r1> B<,> I<r2> B<,> I<value>

Store the content a register in a memory cell.
I<label> is a data label and I<value> is a signed integer.

Memory[I<r1>] = I<r2>

Memory[I<label>] = I<r2>

Memory[I<r1> + I<value>] = I<r2>

=cut

$A{store} = "RR";
sub vam_store   {$M[cm($R[$_[2]])] = $R[$_[3]];       $R[0]++;}

$A{store_l} = "DR";
sub vam_store_l {$M[cm($_[2])] = $R[$_[3]];           $R[0]++;}

$A{store_c} = "RRI";
sub vam_store_i {$M[cm($R[$_[2]]+$_[4])] = $R[$_[3]]; $R[0]++;}

###############################################################################

=pod

=item B<istore> I<r1> B<,> I<r2> B<,> I<r3>

Indirect store with register offset:
Memory[Memory[I<r1>] + I<r2>] =  I<r3>

=cut

$A{istore} = "RRR";
sub vam_istore {$M[cm($M[cm($R[$_[2]])] + $R[$_[4]])] = $R[$_[3]]; $R[0]++;}

###############################################################################

=pod

=item B<istore_c> I<r1> B<,> I<r2> B<,> I<value>

Indirect store with constant offset:
Memory[Memory[I<r1>] + I<value>] = I<r2>

I<value> is an (signed) integer number.

=cut

$A{istore_c} = "RRI";
sub vam_istore_c {$M[cm($M[cm($R[$_[2]])] + $_[4])] = $R[$_[3]] ; $R[0]++;}

###############################################################################

=pod

=back

=cut

###############################################################################
###############################################################################

=pod

=head2 Stack Related Instructions

Stack grows from higher to lower addresses.

=over

=item B<push> I<r1> B<,> I<r2>

Push the value of I<r2> to the stack pointed by I<r1>.

I<r1> = I<r1> - 1; Memory[I<r1>] = I<r2>

I<r1> should contain an address, while I<r2> may hold any value.

=cut

$A{push} = "RR";
sub vam_push {$M[cm(--$R[$_[2]])] = $R[$_[3]]; $R[0]++;}

###############################################################################

=pod

=item B<pop> I<r1> B<,> I<value>

=item B<pop_r> I<r1> B<,> I<r2>

Pop I<values> / I<r2>  many values from the stack.

I<r1> = I<r1> + I<r2>

I<r1> = I<r1> + I<value>

I<r1> should contain an address, while I<r2> / I<value> is an integer.

=cut

$A{pop}   = "RI"; sub vam_pop   {$R[$_[2]] += $_[3];     $R[0]++;}
$A{pop_r} = "RR"; sub vam_pop_r {$R[$_[2]] += $R[$_[3]]; $R[0]++;}


###############################################################################

=pod

=back

=cut

###############################################################################
###############################################################################

=pod

=head2 Input/Output Instructions

The VAM has a build in BIOS (basic I/O system) which allows to
write character strings and numbers to a file (see B<--output>).
Only numbers may be read from a file (see B<--input>).

###############################################################################

=pod

=head3 Input

The following instruction read some values from the input file.

=over

=item B<read_i> I<r1> B<,> I<r2>

=item B<read_f> I<r1>B<,> I<r2>

Read a (signed) integer / floating point number and store it in register I<r1>
and store an error code in I<r2>.
If the read value is an integer or floating point constant, then I<r2>
holds the value 1 (true), else 0 (false)

If I<r1> and I<r2> denote the same register, no error code is stored.
In case of an error the value 0 is stored in the register.

The read number must be terminated by RETURN (newline).

=cut

$A{read_i} = "RR";

sub vam_read_i
{
    my $input = <STDIN>;
    if (defined $input && $input =~ /^[+-]?\d+$/) {
	$R[$_[3]] = 1;
	$R[$_[2]] = $input;
    } else {
	$R[$_[2]] = $R[$_[3]] = 0;
    }
    $R[0]++;
}

$A{read_f} = "RR";
sub vam_read_f
{
    my $input = <STDIN>;
    if (defined $input && $input =~ /^[+-]?\d+(\.\d+)?(e[+-]?\d+)?$/i) {
	$R[$_[3]] = 1;
	$R[$_[2]] = $input;
    } else {
	$R[$_[2]] = $R[$_[3]] = 0;
    }
    $R[0]++;
}

###############################################################################

=pod

=item eof I<r>

Test if End Of File (EOF) has been reached: I<r> = EOF? 1 : 0 (in C speak)

=cut

$A{eof} = "R";
sub vam_eof {$R[$_[2]] = (eof STDIN) ? 1 : 0; $R[0]++;}

###############################################################################

=pod

=back

=cut

###############################################################################
###############################################################################

=pod

=head3 Output

The following instruction write some values to the output file.

=over

=item B<write_s> I<label>

Write the character string denoted by I<label>.

=cut

$A{write_s} = "S";
sub vam_write_s
{
    my $str = $S{$_[2]};
    $str =~ s/\\n/\n/g;
    $str =~ s/\\t/\t/g;
    printf "%s", $str;
    $R[0]++;
}

###############################################################################

=pod

=item B<write_i> I<R>

Write the integer value stored in register I<R>.

=cut

$A{write_i} = "R";
sub vam_write_i {printf "%d", $R[$_[2]]; $R[0]++;}

###############################################################################

=pod

=item B<write_f> I<R>

Write the floating point value stored in register I<R>.

=back

=cut

$A{write_f} = "R";
sub vam_write_f {printf "%f", $R[$_[2]]; $R[0]++;}

###############################################################################


###############################################################################
## Map instruction text to subroutines
###############################################################################

sub get_args($\@$)
{
    my $instr    = shift; # (a hash value of @I)
    my $words    = shift; #
    my $expected = shift; # expected arguments,
                          # for each argument one character:
                          # R : register
                          # L : instruction label
                          # S : string label
                          # D : data label
                          # I : integer constant
                          # F : float constant
    my @args = ();
    my $cnt  = 0;
    for my $s (split (//, $expected)) {
	$cnt++;
	my $a = (defined $words->[$cnt]) ? $words->[$cnt] : "";
	if ($s eq "R") {
	    # Register:  R<number>
	    if ($a =~ /^r(\d+)$/i) {
		$a = $1;
	    } else {
		$a = -1;
	    }
	    (0 <= $a && $a < $max_registers) ||
		syntax ($instr,
			"Expected register name R0 ... ", $max_registers-1);
	} elsif ($s eq "L") {
	    exists ($L{$a}) ||
		syntax ($instr, "Instruction label `$a' not declared");
	    $a = $L{$a};
	} elsif ($s eq "S") {
	    exists ($S{$a}) ||
		syntax ($instr, "String label `$a' not declared");
	} elsif ($s eq "D") {
	    exists ($D{$a}) ||
		syntax ($instr, "Data label `$a' not declared");
	    $a = $D{$a}{address};
	} elsif ($s eq "I") {
	    ($a =~ /^[+-]?\d+$/) ||
		syntax ($instr, "Integer value expected");
	} elsif ($s eq "F") {
	    ($a =~ /^[+-]?\d+(\.\d+)?(e[+-]?\d+)?$/i) ||
		syntax ($instr, "Floating point value expected");
        } else {
	    error "get_args failed";
	}

	push @args, $a;
    }
    $#{@$words} == $cnt ||
	syntax ($instr, "Wrong number of instruction arguments. Given: ",
		$#{@$words}, " expected: $cnt");

    return @args;
}

#############################################################################

sub map_instr
# map the parsed instructions to function calls, check parameters of the
# instructions
{
    ($last_instr < $max_memory) ||
	syntax2 ("", 0, "Memory too small, too many instructions\n");

    for my $i (@I) {
	my @words = split (/[\s,]+/, $i->{text});
	# printf "xxx %s\n", join ("|", @words);
	my $name = lc ($words[0]); # instruction name
	if ($name eq "") {
	    $name  = $i->{text} = "nop";
	    @words = ["nop"];
	}
	if (exists ($A{$name})) {
	    my @args = get_args ($i, @words, $A{$name});
	    $M[$i->{address}] = $i->{instr} = [\&{"vam_$name"}, $i, @args];
	} else {
	    syntax ($i, "Unknown instruction: $name");
	}
    }
}

###############################################################################

sub map_data
# allocate memory for the global data declarations.
{
    $malloc_start = $last_instr+1;
    for my $label (sort {$a cmp $b} keys %D) {
	$D{$label}{address} = $malloc_start;
	$malloc_start += $D{$label}{size};
    }
}

###############################################################################
## Interprete assembler statements
###############################################################################

sub interprete()
{
    # printf "interprete....\n";
    $R[0] = 1;                # instruction counter
    $R[1] = $max_memory-1;    # stack pointer
    $R[2] = $malloc_start;    # free memory pointer
    my $cnt = 0;

    for (my $i = 2; $i < $max_registers; $i++) {
	$R[$i] = 0;
    }

    my $start_time = times();
    while ($R[0] > 0 && $R[0] <= $last_instr) {
	$cnt++;
	# printf "\n%3d: `%s' ", $I[$R[0]]{address}, $I[$R[0]]{text},;
	(ref ($M[$R[0]]) eq "ARRAY" && ref ($M[$R[0]][0]) eq "CODE") ||
	 syntax ($I[$R[0]],
		 "Memory address " . $I[$R[0]]{address} .
		 " corrupted; instruction expected");
        &{$M[$R[0]][0]} (@{$M[$R[0]]});
	# printf "\n --> %d\n", $R[0];
    }
    ($R[0] <= $last_instr) ||
	syntax2 ("",0, "instruction counter out of range\n");

    if ($option_statistics) {
	my $end_time = times();
	printf "\nSome statistics:\n";
	printf "instructions:          %7d\n", $last_instr;
	printf "executed instructions: %7d\n", $cnt;
	printf "user time              %7.3f seconds\n",
	       $end_time - $start_time;
    }
}

###############################################################################
## Parse assembler source
###############################################################################

sub parse_line ($)
{
    my $line = shift;
    # printf "parse: <$line>\n";

    $line =~ s/#.*$//;       # remove comments
    $line =~ s/^\s*//g;      # remove lading and ...
    $line =~ s/\s*$//g;      # ... trailing blanks
    # TODO: strings, which contain #
    return if ($line =~ /^\s*$/); # skip empty lines

    # extract label (if given)
    my $label = "";
    if ($line =~ /^((\w[\w\d_]*)\s*:\s*)(.*)$/) {
	$label = $2;
	$line  = $3;
    }

    # printf "  label=<$label> line=<$line>\n";

    if ($line =~ /^"(.*)"$/) {
	# a string declaration
	# add it to the string pool
	$label     || syntax2 ($ARGV, $NR,
			       "A string declaration needs a label");
	$S{$label} && syntax2 ($ARGV, $NR,
			       "String label `$label' declared more than once");
	$S{$label} = $1;
	return;
    } elsif ($line =~ /^DATA\s+(\d+)$/) {
	# a data declaration
	$label     || syntax2 ($ARGV, $NR,
			       "A data declaration needs a label");
	$D{$label} && syntax2 ($ARGV, $NR,
			       "Data label `$label' declared more than once");
	$D{$label}{size}    = $1;
	$D{$label}{address} = 0; # to be computed.
	return;
    }

    $last_instr++;
    my $instr = {file    => $ARGV,
		 line    => $NR,
		 label   => $label,
		 address => $last_instr,
		 text    => $line};
    $I[$last_instr] = $instr;

    if ($label) {
	$L{$label} && syntax ($instr,
			      "Instruction label `$label' declared more than once");
	$L{$label} = $last_instr;
    }
}

###############################################################################
## check command line options and arguments
###############################################################################

Getopt::Long::Configure ("auto_abbrev", "no_ignore_case",
			 "getopt_compat",  "require_order");

$max_registers = 32;
$max_memory    = "32M";

GetOptions (
   "help|h"      => sub {pod2usage (-exitval => 1, -verbose => 1);},
   "manual|M"    => sub {pod2usage (-exitval => 1, -verbose => 2);},
   "verbose"     => \$option_verbose,
   "Version"     => sub {printf "$CMD: Version $VERSION\n"; exit;},
   "input=s"     => \$option_input,
   "output=s"    => \$option_output,
   "dump|D"      => \$option_dump,
   "statistics"  => \$option_statistics,
   "registers=i" => \$max_registers,
   "memory|m=s"  => \$max_memory,
   ) or pod2usage(-exitval => 2, -verbose => 0,
		  -message => "Try --help or --manual");

if ($max_memory =~ /^(\d*)([mMkK])$/) {
    $max_memory = $1 * 1024;
    $max_memory *= 1024 if ($2 eq "M" || $2 eq "m");
} else {
    error "Wrong argument for --memory: $max_memory";
}

($max_registers >  3)    || error "--registers: at least 4 registers needed";
($max_memory    >= 1024) || error "--memory: at least 1K must be given";


###############################################################################
## start processing
###############################################################################

# initialize
$| = 1;

# Memory[0] = I[0] = "undefined"
$last_instr = 0;
$I[0] = {file    => "",
	 line    => 0,
	 label   => "",
	 address => 0,
	 text    => "nop"};

# parse source files
while (<>) {
    chomp;
    parse_line ($_);
}

# fix addresses and map instruction text to subroutines
map_data();
map_instr();

# debugging
dump_memory() if ($option_dump);

# set up input and output of the program
if ($option_input) {
    open (IN, "< $option_input") ||
	error ("can not open $option_input for reading");
    *STDIN = *IN;
}

if ($option_output) {
    open (OUT, "> $option_output") ||
	error ("can not open $option_output for writing");
    *STDIN = *IN;
}

# and finally do the job:
interprete();

# clean up
close (IN)  if ($option_input);
close (OUT) if ($option_output);

# we're done.
exit;

###############################################################################

=pod

=head1 EXAMPLES

The following program prints the square numbers 1 ... n^2:

 #############################################################
 # Description:  A VAM program computing square numbers
 #               Input:   read a number
 #               Output:  write square numbers 1 .... n^2

 s_in:  "Please input an integer: "
 s_out: "^2 = "
 NL:    "\n"


 start:
       write_s s_in
       read_i  R4, R4             # R4: n (ignore errors)
       write_s NL

       cload_i R5, 1	          # R5: i = 1

 loop:
       cmp_i   R6, R5, R4         # if (i > n) goto end
       ifgt    R6, end

       mult_i  R6, R5, R5         # i * i

       write_i R5                 # printf ("%d^2 = %d\n", i, i*i);
       write_s s_out
       write_i R6
       write_s NL

       add_c   R5, R5, 1          # i = i + 1

       goto loop                  # and loop

 end:
       end

 #############################################################

More examples will be found in the example files shipped together with this
program.

=head1 REQUIREMENTS

B<perl>(1)

=head1 AUTHOR

=for latex
{\setlength{\parindent}{0pt}

Dr. Jrgen Vollmer <Juergen.Vollmer@informatik-vollmer.de>

Copyright (C) 2005 Dr. Jrgen Vollmer, Karlsruhe.

Homepage of B<VAM>: http://www.informatik-vollmer.de/software/vam.html

If have written "large programs" doing some interesting job, it would be nice to
send me your VAM source.

If you find this software useful, I would be glad to receive a postcard from
you, showing the place where you're living:

=for latex
}

Dr. Jrgen Vollmer, Viktoriastrasse 15,
D-76133 Karlsruhe, Germany.


=head1 LICENSE

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

=for latex
\noindent\url{http://www.gnu.org/copyleft/gpl.html}

=for man
http://www.gnu.org/copyleft/gpl.html

=for html
http://www.gnu.org/copyleft/gpl.html



=cut

###############################################################################

=pod

=head1 VERSION

Version 1.0 of 2005/02/24

=head1 CHANGELOG

  $Log: vam.pl,v $
  Revision 1.8  2005/07/12 12:06:43  vollmer
  *** empty log message ***

  Revision 1.7  2005/07/12 11:20:57  vollmer
  *** empty log message ***

  Revision 1.6  2005/06/19 10:41:10  vollmer
  fixed focu

  Revision 1.5  2005/03/11 22:58:51  vollmer
  fixed name of vam_ifne

  Revision 1.4  2005/03/11 22:51:40  vollmer
  sub cm: if a memory cell undefined during an access, assign 0 to it

  Revision 1.3  2005/03/11 22:48:05  vollmer
  fixed sub_f

  Revision 1.2  2005/03/11 14:11:45  vollmer
  fixed docu

  Revision 1.1  2005/02/24 16:55:52  vollmer
  Initial revision


=cut


###############################################################################
